/*
 *  This file is part of BGX, the Bayesian Gene eXpression program.
 *  Copyright (c) 2003-2004  Graeme Ambler <graeme@ambler.me.uk>
 *
 *  BGX is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License, version 2, as
 *  published by the Free Software Foundation.
 *
 *  BGX is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "rwm.hh"

#define SAFE_EXP(a) exp(std::max(-500.0,std::min(0.0,a)))

inline double pnorm(double x){ return .5+.5*erf(M_SQRT1_2*x); }

class S_T{
private:
  array2d& PM;
  array2d& MM;
  array2d& H;
  double& phi;
  array2d& mu;
  array2d& taug;
  array& tau;
  array& beta;
  int* probesets;
  int* samplesets;
  int g;
  int p;
  int c;
  int r;
public:
  S_T(array2d& PM_, array2d& MM_, array2d& H_, double& phi_, array2d& mu_, 
      array2d& taug_, array& tau_, array& beta_, int* probesets_,
      int* samplesets_)
    : PM(PM_), MM(MM_), H(H_), phi(phi_), mu(mu_), taug(taug_), 
      tau(tau_), beta(beta_), probesets(probesets_), samplesets(samplesets_) {}

  double operator()(array2d& S, double S2, int j, int i)
  {
    if(i==0){
      g=p=0;
      if(j==0) c=r=0;
      else if(++r==samplesets[c]) ++c,r=0;
    }
    else if(++p==probesets[g]) ++g,p=0;
    if(S2>0)
      {
	double x=log(S[j][i]+1);
	double x2=log(S2+1);
	return SAFE_EXP(taug[c][g]*((x-mu[c][g])*(x-mu[c][g])-(x2-mu[c][g])*(x2-mu[c][g]))*.5
			+ (x-x2)
			+ tau[j]*((PM[j][i]-S[j][i]-H[j][i]-beta[j])*(PM[j][i]-S[j][i]-H[j][i]-beta[j]) 
			       + (MM[j][i]-phi*S[j][i]-H[j][i]-beta[j])*(MM[j][i]-phi*S[j][i]-H[j][i]-beta[j])
			       - (PM[j][i]-S2-H[j][i]-beta[j])*(PM[j][i]-S2-H[j][i]-beta[j]) 
			       - (MM[j][i]-phi*S2-H[j][i]-beta[j])*(MM[j][i]-phi*S2-H[j][i]-beta[j]))*.5 );
      }
    else return 0;
  }
};

class H_T{
private:
  array2d& PM;
  array2d& MM;
  array2d& S;
  double& phi;
  array& lambda;
  array& eta;
  array& tau;
  array& beta;
public:
  H_T(array2d& PM_, array2d& MM_, array2d& S_, double& phi_, array& lambda_, 
      array& eta_, array& tau_, array& beta_)
    : PM(PM_), MM(MM_), S(S_), phi(phi_), lambda(lambda_), eta(eta_), 
      tau(tau_), beta(beta_) {}

  double operator()(array2d& H, double H2, int j, int i)
  {
    if(H2>0)
      {
	double y=log(H[j][i]+1);
	double y2=log(H2+1);
	return SAFE_EXP(eta[j]*( (y-lambda[j])*(y-lambda[j]) - 
			       (y2-lambda[j])*(y2-lambda[j]) )*.5
			+ (y-y2)
			+ tau[j]*((PM[j][i]-S[j][i]-H[j][i]-beta[j])*(PM[j][i]-S[j][i]-H[j][i]-beta[j]) 
			       + (MM[j][i]-phi*S[j][i]-H[j][i]-beta[j])*(MM[j][i]-phi*S[j][i]-H[j][i]-beta[j])
			       - (PM[j][i]-S[j][i]-H2-beta[j])*(PM[j][i]-S[j][i]-H2-beta[j]) 
			       - (MM[j][i]-phi*S[j][i]-H2-beta[j])*(MM[j][i]-phi*S[j][i]-H2-beta[j]))*.5 );
      }
    else return 0;
  }
};

class Mu_T{
private:
  array2d& S;
  array2d& taug;
  int* probesets;
  int* samplesets;
  int offset;
  int offset2;
public:
  Mu_T(array2d& S_, array2d& taug_, int* probesets_, int* samplesets_)
    : S(S_), taug(taug_), probesets(probesets_), samplesets(samplesets_) {}

  double operator()(array2d& mu, double mu2, int k, int i)
  {
    if(i==0){
      offset=0;
      if(k==0) offset2=0;
      else offset2 += samplesets[k-1];
    }
    else offset += probesets[i-1];
    if(mu2>0 && mu2<15)
      {
	double sum=0;
	for(int l=0; l<samplesets[k]; ++l)
	  {
	    for(int j=0; j<probesets[i]; ++j)
	      {
		double x=log(S[offset2+l][offset+j]+1);
		sum += (x-mu[k][i])*(x-mu[k][i]) - (x-mu2)*(x-mu2);
	      }
	  }
	return SAFE_EXP(taug[k][i]*( sum )*.5 + 
			samplesets[k]*probesets[i]*log(pnorm(mu[k][i]*sqrt(taug[k][i]))/
						       pnorm(mu2*sqrt(taug[k][i]))));
      }
    else return 0;
  }
};

class TauG_T{
private:
  array2d& S;
  array2d& mu;
  array& a;
  array& b;
  int* probesets;
  int* samplesets;
  int offset;
  int offset2;
public:
  TauG_T(array2d& S_, array2d& mu_, array& a_, array& b_, int* probesets_, int* samplesets_)
    : S(S_), mu(mu_), a(a_), b(b_), probesets(probesets_), samplesets(samplesets_) {}

  double operator()(array2d& taug, double taug2, int k, int i)
  {
    if(i==0){
      offset=0;
      if(k==0) offset2=0;
      else offset2 += samplesets[k-1];
    }
    else offset += probesets[i-1];
    if(taug2>0)
      {
	double sum=0;
	for(int l=0; l<samplesets[k]; ++l)
	  {
	    for(int j=0; j<probesets[i]; ++j)
	      {
		double x=log(S[offset2+l][offset+j]+1);
		sum += (x-mu[k][i])*(x-mu[k][i]);
	      }
	  }
	return 
	  SAFE_EXP((taug[k][i]-taug2)*( sum )*.5
		   + (samplesets[k]*probesets[i]*.5-1)*log(taug2/taug[k][i])
		   + samplesets[k]*probesets[i]*log(pnorm(mu[k][i]*sqrt(taug[k][i]))/
			       pnorm(mu[k][i]*sqrt(taug2)))
		   + b[k]*((log(taug[k][i])-a[k])*(log(taug[k][i])-a[k])
			   -(log(taug2)-a[k])*(log(taug2)-a[k]))*.5);
      }
    else return 0;
  }
};


class Lambda_T{
private:
  array2d& H;
  array& eta;
  double p_mu;
  double p_tau;
  int probes;
public:
  Lambda_T(array2d& H_, array& eta_, double p_mu_, double p_tau_)
    : H(H_), eta(eta_), p_mu(p_mu_), p_tau(p_tau_), probes(H[0].size()) {}

  double operator()(array& lambda, double lambda2, int j)
  {
    double sum=0;
    for(int i=0; i<probes; ++i)
      {
	double yi=log(H[j][i]+1);
	sum += (yi-lambda[j])*(yi-lambda[j]) - (yi-lambda2)*(yi-lambda2);
      }
    return SAFE_EXP(eta[j]*sum*.5 + probes*log(pnorm(lambda[j]*sqrt(eta[j]))/
					       pnorm(lambda2*sqrt(eta[j])))
		    + p_tau*( (lambda[j]-p_mu)*(lambda[j]-p_mu)
			      - (lambda2-p_mu)*(lambda2-p_mu) )*.5 );
  }
};

class Eta_T{
private:
  array2d& H;
  array& lambda;
  double alpha;
  double beta;
  int probes;
public:
  Eta_T(array2d& H_, array& lambda_, double alpha_, double beta_)
    : H(H_), lambda(lambda_), alpha(alpha_), beta(beta_), probes(H[0].size()) {}

  double operator()(array& eta, double eta2, int j)
  {
    double sum=0;
    for(int i=0; i<probes; ++i)
      {
	double yi=log(H[j][i]+1);
	sum += (yi-lambda[j])*(yi-lambda[j]);
      }
    return SAFE_EXP((eta[j]-eta2)*(.5*sum+beta)
		    + probes*log(pnorm(lambda[j]*sqrt(eta[j]))/
				 pnorm(lambda[j]*sqrt(eta2)))
		    + (alpha+.5*probes-1)*log(eta2/eta[j]));
  }
};

class Phi_T{
private:
  double& phi;
  array2d& MM;
  array2d& S;
  array2d& H;
  array& tau;
  array& beta;
  Random* rand;
  int probes;
  int samples;
public:
  Phi_T(double& phi_, array2d& MM_, array2d& S_, array2d& H_, array& tau_, array& beta_,
	Random* rand_)
    : phi(phi_), MM(MM_), S(S_), H(H_), tau(tau_), beta(beta_), rand(rand_), probes(MM[0].size()), samples(MM.size()) {}

  void Update()
  {
    double sumMMmHTot=0;
    double sumSTot=0;
    for(int j=0; j<samples; ++j)
      {
	double sumMMmH=0;
	double sumS=0;
	for(int i=0; i<probes; ++i)
	  {
	    sumMMmH += S[j][i]*(MM[j][i]-H[j][i]-beta[j]);
	    sumS += S[j][i]*S[j][i];
	  }
	sumMMmHTot += sumMMmH*tau[j];
	sumSTot += sumS*tau[j];
      }
    phi=rand->Normal( sumMMmHTot/sumSTot, sqrt(1.0/(sumSTot)) );
  }
};

class Tau_T{
private:
  array& tau;
  array2d& PM;
  array2d& MM;
  array2d& S;
  array2d& H;
  double& phi;
  array& beta;
  double alpha;
  double p_beta;
  Random* rand;
  int probes;
  int samples;
public:
  Tau_T(array& tau_, array2d& PM_, array2d& MM_, array2d& S_, array2d& H_, 
	double& phi_, array& beta_, double alpha_, double p_beta_, Random* rand_)
    : tau(tau_), PM(PM_), MM(MM_), S(S_), H(H_), phi(phi_), beta(beta_), alpha(alpha_), 
      p_beta(p_beta_), rand(rand_), probes(PM[0].size()), samples(PM.size()) {}

  void Update()
  {
    for(int j=0; j<samples; ++j)
      {
	double sum=0;
	for(int i=0; i<probes; ++i)
	  {
	    sum += ( ( PM[j][i]-(S[j][i]+H[j][i]+beta[j]) ) * ( PM[j][i]-(S[j][i]+H[j][i]+beta[j]) ) + 
		 ( MM[j][i]-(phi*S[j][i]+H[j][i]+beta[j]) ) * ( MM[j][i]-(phi*S[j][i]+H[j][i]+beta[j]) ) );
	  }
	tau[j]=rand->Gamma( alpha+probes, p_beta+sum*.5 );
      }
  }
};

class A_T{
private:
  array& a;
  array2d& taug;
  array& b;
  double p_mu;
  double p_tau;
  Random* rand;
  int genes;
  int conditions;
public:
  A_T(array& a_, array2d& taug_, array& b_,
      double p_mu_, double p_tau_, Random* rand_)
    : a(a_), taug(taug_), b(b_),
      p_mu(p_mu_), p_tau(p_tau_), rand(rand_), genes(taug[0].size()), conditions(taug.size()) {}

  void Update()
  {
    for(int j=0; j<conditions; ++j)
      {
	double sum=0;
	for(int i=0; i<genes; ++i)
	  {
	    sum += log(taug[j][i]);
	  }
	a[j]=rand->Normal( (p_tau*p_mu+b[j]*sum)/(p_tau+b[j]*genes), 
			   sqrt( 1.0/(p_tau+b[j]*genes) ) );
      }
  }
};

class B_T{
private:
  array& b;
  array2d& taug;
  array& a;
  double alpha;
  double beta;
  Random* rand;
  int genes;
  int conditions;
public:
  B_T(array& b_, array2d& taug_, array& a_,
      double alpha_, double beta_, Random* rand_)
    : b(b_), taug(taug_), a(a_),
      alpha(alpha_), beta(beta_), rand(rand_), genes(taug[0].size()), conditions(taug.size()) {}

  void Update()
  {
    for(int j=0; j<conditions; ++j)
      {
	double rss=0;
	for(int i=0; i<genes; ++i)
	  {
	    rss += (log(taug[j][i])-a[j])*(log(taug[j][i])-a[j]);
	  }
	b[j]=rand->Gamma( alpha+genes*.5, beta+rss*.5 );
      }
  }
};

class Beta_T{
private:
  array& beta;
  array2d& PM;
  array2d& MM;
  array2d& S;
  array2d& H;
  double& phi;
  array& tau;
  double mu;
  double sigma;
  Random* rand;
  int probes;
  int samples;
public:
  Beta_T(array& beta_, array2d& PM_, array2d& MM_, array2d& S_, array2d& H_,
      double& phi_, array& tau_, double mu_, double sigma_, Random* rand_)
    : beta(beta_), PM(PM_),  MM(MM_), S(S_), H(H_), phi(phi_), tau(tau_),
      mu(mu_), sigma(sigma_), rand(rand_), probes(PM[0].size()), samples(PM.size()) {}

  void Update()
  {
    for(int j=1; j<samples; ++j)
      {
	double sum=0;
	for(int i=0; i<probes; ++i)
	  {
	    sum += PM[j][i]-S[j][i]-2*H[j][i]+MM[j][i]-phi*S[j][i];
	  }
	beta[j]=rand->Normal( tau[j]*sum/(probes*tau[j]+1/(2*sigma)), sqrt(1.0/(probes*tau[j]+1/(2*sigma))) );
      }
  }
};
