/*
 *  This file is part of BGX, the Bayesian Gene eXpression program.
 *  Copyright (c) 2003-2004  Graeme Ambler <graeme@ambler.me.uk>
 *
 *  BGX is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License, version 2, as
 *  published by the Free Software Foundation.
 *
 *  BGX is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifndef LOCAL_RWM_INCLUDED
#define LOCAL_RWM_INCLUDED

#include "rand.hh"

template<class Accept, typename Update_T=std::valarray<double>, class Generator=Random>
class RWM{
 private:
  Update_T& Param;
  Accept& AccProb;
  double jmp_size;
  Generator* rand;
  Update_T nAccept;

  int length;
 public:
  RWM(Update_T& Param_, Accept& AccProb_, double jmp_size_, Generator* rand_)
    : Param(Param_), AccProb(AccProb_), jmp_size(jmp_size_), rand(rand_), nAccept(Param)
  { nAccept=0; length=Param.size(); }

  void Update()
  {
    for(int i=0; i<length; ++i)
      {
	double proposal=rand->Normal(Param[i],jmp_size);
	if(rand->Uniform() < AccProb(Param,proposal,i))
	  {
	    Param[i]=proposal;
	    ++nAccept[i];
	  }
      }
  }

  void Reset() { for(int i=0; i<length; ++i) nAccept[i]=0; }

  void pAccept(int nTry, double* pAcc)
  {
    for(int i=0; i<length; ++i)
      pAcc[i] = nAccept[i]/(double) nTry;
  }
};

template<class Accept, class Generator>
class RWM<Accept,std::valarray<std::valarray<double> >,Generator>{
 private:
  std::valarray<std::valarray<double> >& Param;
  Accept& AccProb;
  double jmp_size;
  Generator* rand;
  std::valarray<std::valarray<double> > nAccept;

  int length;
 public:
  RWM(std::valarray<std::valarray<double> >& Param_, Accept& AccProb_, 
      double jmp_size_, Generator* rand_)
    : Param(Param_), AccProb(AccProb_), jmp_size(jmp_size_), rand(rand_), 
      nAccept(Param)
  {
    length=Param.size();
    for(int i=0; i<length; ++i)
      {
	nAccept[i].resize(Param[i].size());
	nAccept[i]=0;
      }
  }

  void Update()
  {
    for(int j=0; j<length; ++j)
      {
	for(unsigned int i=0; i<Param[j].size(); ++i)
	  {
	    double proposal=rand->Normal(Param[j][i],jmp_size);
	    if(rand->Uniform() < AccProb(Param,proposal,j,i))
	      {
		Param[j][i]=proposal;
		++nAccept[j][i];
	      }
	  }
      }
  }

  void Reset()
  {
    for(int j=0; j<length; ++j)
      for(unsigned int i=0; i<Param[j].size(); ++i)
	nAccept[j][i]=0;
  }

  void pAccept(int nTry, double* pAcc)
  {
    int k=0;
    for(int i=0; i<length; ++i)
      for(unsigned int j=0; j<Param[i].size(); ++j)
	pAcc[k++] = nAccept[i][j]/(double) nTry;
  }
};

template<class Accept, class Generator>
class RWM<Accept,double,Generator>{
private:
  double& Param;
  Accept& AccProb;
  double jmp_size;
  Generator* rand;
  double nAccept;
public:
  RWM(double& Param_, Accept& AccProb_, double jmp_size_, Generator* rand_)
    : Param(Param_), AccProb(AccProb_), jmp_size(jmp_size_), rand(rand_), nAccept(Param)
  { nAccept=0; }

  void Update()
  {
    double proposal=rand->Normal(Param,jmp_size);
    if(rand->Uniform() < AccProb(Param,proposal))
      {
	Param=proposal;
	++nAccept;
      }
  }

  void Reset(){ nAccept=0; }

  void pAccept(int nTry, double* pAcc)
  {
    *pAcc = nAccept/(double) nTry;
  }
};

#endif // LOCAL_RWM_INCLUDED
